/*
 * Plasma Screen Saver
 * By Carl Olsson
 *
 * Screen saver code based on wintests\scrsave.c from Allegro.
 * Plasma code based on plasma by Plasma by Jan Mller & Erik Hansen.
 */

#include <time.h>
#include <math.h>
#include <stdlib.h>
#include <allegro.h>
#include <winalleg.h>

#ifndef PI
   #define PI 3.141592654
#endif

#define SCREEN_WIDTH    320
#define SCREEN_HEIGHT   240

#define VIEW_WIDTH      SCREEN_WIDTH
#define VIEW_HEIGHT     SCREEN_HEIGHT

#define TABLE_WIDTH     (VIEW_WIDTH * 2)
#define TABLE_HEIGHT    (VIEW_HEIGHT * 2)

#define PLASMA_MAGNITUDE (16.0 / 1.0)
#define PLASMA_CONTRAST  (1.0 / 64.0)

#define DELTA_RANGE 0.25

//#define mycol(u, a) (unsigned char)((cos((u) + (a)) + 1) * 31) // this line or...
#define mycol(u, a) (unsigned char)(MAX(0.0, cos((u) + (a))) * 63) // ...try this line instead

typedef struct PNT {
   float x;
   float y;
} PNT;

BITMAP *buf;

PNT point1, point2, point3, point4;
PNT p1, p2, p3, p4;
float delta1x, delta1y, delta2x, delta2y, delta3x, delta3y, delta4x, delta4y;

unsigned char table_1[TABLE_WIDTH * TABLE_HEIGHT];

unsigned char table_siner(float dist)
{
   return (unsigned char)((sin(dist / PLASMA_MAGNITUDE)) / PLASMA_CONTRAST);
}

unsigned char table_straight(float dist)
{
   return (unsigned char)(dist / PLASMA_MAGNITUDE / PLASMA_CONTRAST);
}

/* generates a table based on a function */
void calc_table(unsigned char table[TABLE_WIDTH * TABLE_HEIGHT], unsigned char (*function)(float dist))
{
   int _x, _y;
   float dist;

   for (_y = 0; _y < TABLE_HEIGHT; _y++) {
      for (_x = 0; _x < TABLE_WIDTH; _x++) {
         dist = sqrt(pow((TABLE_HEIGHT / 2) - _y, 2) + pow((TABLE_WIDTH / 2) - _x, 2));
         table[(_y * TABLE_WIDTH) + _x] = function(dist);
      }
   }
}

/* updates the palette */
void calc_colours(void)
{
   static double r = 1.0 / 6.0 * PI,
                 g = 3.0 / 6.0 * PI,
                 b = 5.0 / 6.0 * PI;
   int i = 0;
   double u;
   RGB colour;

   while (i < 256) {
      u = 2 * PI / 256 * i;
      colour.r = mycol(u, r);
      colour.g = mycol(u, g);
      colour.b = mycol(u, b);
      set_color(i, &colour);
      i++;
   }
   r += 0.05;
   g -= 0.05;
   b += 0.1;
}

/* initialises our graphical effect */
void ss_init()
{
   calc_table(table_1, table_siner);

   point1.x = 0.0;
   point1.y = 0.0;
   point2.x = 0.0;
   point2.y = 0.0;
   point3.x = 0.0;
   point3.y = 0.0;
   point4.x = 0.0;
   point4.y = 0.0;

   p1.x = 0.0;
   p1.y = 0.0;
   p2.x = 0.0;
   p2.y = 0.0;
   p3.x = 0.0;
   p3.y = 0.0;
   p4.x = 0.0;
   p4.y = 0.0;

   delta1x = ((float)rand() / RAND_MAX) * DELTA_RANGE - (DELTA_RANGE / 2);
   delta1y = ((float)rand() / RAND_MAX) * DELTA_RANGE - (DELTA_RANGE / 2);
   delta2x = ((float)rand() / RAND_MAX) * DELTA_RANGE - (DELTA_RANGE / 2);
   delta2y = ((float)rand() / RAND_MAX) * DELTA_RANGE - (DELTA_RANGE / 2);
   delta3x = ((float)rand() / RAND_MAX) * DELTA_RANGE - (DELTA_RANGE / 2);
   delta3y = ((float)rand() / RAND_MAX) * DELTA_RANGE - (DELTA_RANGE / 2);
   delta4x = ((float)rand() / RAND_MAX) * DELTA_RANGE - (DELTA_RANGE / 2);
   delta4y = ((float)rand() / RAND_MAX) * DELTA_RANGE - (DELTA_RANGE / 2);
}



/* animates the graphical effect */
void ss_update()
{
   point1.x += delta1x;
   point1.y += delta1y;
   point2.x += delta2x;
   point2.y += delta2y;
   point3.x += delta3x;
   point3.y += delta3y;
   point4.x += delta4x;
   point4.y += delta4y;
   point1.x = fmod(point1.x, VIEW_WIDTH);
   point1.y = fmod(point1.y, VIEW_HEIGHT);
   point2.x = fmod(point2.x, VIEW_WIDTH);
   point2.y = fmod(point2.y, VIEW_HEIGHT);
   point3.x = fmod(point3.x, VIEW_WIDTH);
   point3.y = fmod(point3.y, VIEW_HEIGHT);
   point4.x = fmod(point4.x, VIEW_WIDTH);
   point4.y = fmod(point4.y, VIEW_HEIGHT);
   p1.x = (VIEW_WIDTH / 2) + (VIEW_WIDTH / 2) * sin(point1.x);
   p1.y = (VIEW_HEIGHT / 2) + (VIEW_HEIGHT / 2) * sin(point1.y);
   p2.x = (VIEW_WIDTH / 2) + (VIEW_WIDTH / 2) * sin(point2.x);
   p2.y = (VIEW_HEIGHT / 2) + (VIEW_HEIGHT / 2) * sin(point2.y);
   p3.x = (VIEW_WIDTH / 2) + (VIEW_WIDTH / 2) * sin(point3.x);
   p3.y = (VIEW_HEIGHT / 2) + (VIEW_HEIGHT / 2) * sin(point3.y);
   p4.x = (VIEW_WIDTH / 2) + (VIEW_WIDTH / 2) * sin(point4.x);
   p4.y = (VIEW_HEIGHT / 2) + (VIEW_HEIGHT / 2) * sin(point4.y);
   calc_colours();
}



/* draws the graphical effect */
void ss_draw()
{
   int _x, _y;
   char colour;

   clear(buf);

   for (_y = 0; _y < VIEW_HEIGHT; _y++) {
      for (_x = 0; _x < VIEW_WIDTH; _x++) {
         colour = table_1[TABLE_WIDTH * (_y + (int)p1.y) + (_x + (int)p1.x)] +
                  table_1[TABLE_WIDTH * (_y + (int)p2.y) + (_x + (int)p2.x)] +
                  table_1[TABLE_WIDTH * (_y + (int)p3.y) + (_x + (int)p3.x)] +
                  table_1[TABLE_WIDTH * (_y + (int)p4.y) + (_x + (int)p4.x)];
         putpixel(buf, _x, _y, colour);
      }
   }
}



/* shuts down the graphical effect */
void ss_exit()
{
}



/* dialog procedure for the settings dialog */
BOOL CALLBACK settings_dlg_proc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
   switch (uMsg) {

      case WM_COMMAND: 
	 switch (HIWORD(wParam)) { 

	    case BN_CLICKED:
	       EndDialog(hwndDlg, 0);
	       return 1;
	 }
	 break;
   }

   return 0;
}



/* the settings dialog function */
int do_settings(HANDLE hInstance, HANDLE hPrevInstance, HWND hParentWnd)
{
   DialogBox(hInstance, "ID_CONFIG_DLG", hParentWnd, settings_dlg_proc);

   return 0;
}



/* the password dialog function */
int do_password(HANDLE hInstance, HANDLE hPrevInstance, HWND hParentWnd)
{
   MessageBox(hParentWnd, "Sorry, this screensaver doesn't implement the password stuff", "Allegro Screensaver", MB_OK);

   return 0;
}



/* window procedure for the screensaver preview */
long FAR PASCAL preview_wnd_proc(HWND hwnd, UINT message, UINT wParam, LONG lParam)
{
   PAINTSTRUCT ps;
   HDC hdc;

   switch (message) {

      case WM_CREATE:
	 SetTimer(hwnd, 1, 15, NULL);
	 return 0;

      case WM_TIMER:
	 ss_update();
	 InvalidateRect(hwnd, NULL, FALSE);
	 return 0;

      case WM_PAINT:
	 hdc = BeginPaint(hwnd, &ps);
	 ss_draw();
	 set_palette_to_hdc(hdc, _current_palette);
	 draw_to_hdc(hdc, buf, 0, 0);
	 EndPaint(hwnd, &ps);
	 return 0;

      case WM_DESTROY:
	 KillTimer(hwnd, 1);
	 PostQuitMessage(0);
	 return 0;
   }

   return DefWindowProc(hwnd, message, wParam, lParam);
}



/* the screensaver preview function */
int do_preview(HANDLE hInstance, HANDLE hPrevInstance, HWND hParentWnd)
{
   WNDCLASS wndclass;
   HWND hwnd;
   MSG msg;
   RECT rc;

   if (!hPrevInstance) {
      wndclass.style = CS_HREDRAW | CS_VREDRAW;
      wndclass.lpfnWndProc = preview_wnd_proc;
      wndclass.cbClsExtra = 0;
      wndclass.cbWndExtra = 0;
      wndclass.hInstance = hInstance;
      wndclass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
      wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
      wndclass.hbrBackground = NULL;
      wndclass.lpszMenuName = NULL;
      wndclass.lpszClassName = "sspreview";

      RegisterClass(&wndclass);
   }

   if (hParentWnd)
      GetClientRect(hParentWnd, &rc);
   else
      rc.right = rc.bottom = 256;

   install_allegro(SYSTEM_NONE, &errno, atexit);
   set_palette(default_palette);
   set_gdi_color_format();

   buf = create_bitmap(rc.right, rc.bottom);

   ss_init();
   ss_update();

   hwnd = CreateWindow("sspreview", NULL, WS_CHILD, 
		       0, 0, rc.right, rc.bottom, 
		       hParentWnd, NULL, hInstance, NULL);

   ShowWindow(hwnd, SW_SHOW);
   UpdateWindow(hwnd);

   while (GetMessage(&msg, NULL, 0, 0)) {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
   }

   ss_exit();
   destroy_bitmap(buf);

   return msg.wParam;
}



/* run the saver normally, in fullscreen mode */
int do_saver(HANDLE hInstance, HANDLE hPrevInstance, HWND hParentWnd)
{
   int mx, my;

   allegro_init();
   install_keyboard();
   install_mouse();
   install_timer();
   set_color_depth(8);
   set_gfx_mode(GFX_DIRECTX, SCREEN_WIDTH, SCREEN_HEIGHT, 0, 0);

   buf = create_bitmap(SCREEN_W, SCREEN_H);

   ss_init();
   ss_update();

   mx = mouse_x;
   my = mouse_y;

   while ((!keypressed()) && (!mouse_b) && (mouse_x == mx) && (mouse_y == my)) {
      ss_update();
      ss_draw();
      acquire_screen();
      vsync();
      blit(buf, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
      release_screen();
      poll_mouse();
   }

   ss_exit();
   destroy_bitmap(buf);

   return 0;
}



/* the main program body */
int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance, LPSTR lpszCmdParam, int nCmdShow)
{
   HWND hwnd;
   char *args;

   args = lpszCmdParam;

   if ((args[0] == '-') || (args[0] == '/'))
      args++;

   if ((args[0]) && ((args[1] == ' ') || (args[1] == ':')))
      hwnd = (HWND)atoi(args+2);
   else
      hwnd = GetActiveWindow();

   switch (utolower(args[0])) {

      case 'c':
	 return do_settings(hInstance, hPrevInstance, hwnd);

      case 'a':
	 return do_password(hInstance, hPrevInstance, hwnd);

      case 'p':
	 return do_preview(hInstance, hPrevInstance, hwnd);

      case 's':
	 return do_saver(hInstance, hPrevInstance, hwnd);
   }

   return 0;
}


